@@ -1,6 +1,8 @@
 #ifndef INTEGER_COMMON_H
 #define INTEGER_COMMON_H
 
+#include "primetest.h"
+
 //note: gmp already has c++ bindings so could have just used those. oh well
 
 //const bool output_stats=false;
@@ -172,7 +174,7 @@ struct integer {
         res.resize(mpz_sizeinbase(impl, 2)/64 + 1, 0);
 
         size_t count;
-        mpz_export(&res[0], &count, -1, 8, 0, 0, impl);
+        mpz_export(res.data(), &count, -1, 8, 0, 0, impl);
         res.resize(count);
 
         return res;
@@ -331,7 +333,7 @@ struct integer {
     }
 
     bool prime() const {
-        return mpz_probab_prime_p(impl, 50)!=0;
+        return is_prime_bpsw(impl) != 0;
     }
 
     bool operator<(const integer& t) const {
