@@ -1,6 +1,7 @@
 #ifndef PROOF_COMMON_H
 #define PROOF_COMMON_H
 #include "Reducer.h"
+#include "bqfc.c"
 
 std::vector<unsigned char> ConvertIntegerToBytes(integer x, uint64_t num_bytes) {
     std::vector<unsigned char> bytes;
@@ -12,6 +13,7 @@ std::vector<unsigned char> ConvertIntegerToBytes(integer x, uint64_t num_bytes)
     }
     for (int iter = 0; iter < num_bytes; iter++) {
         auto byte = (x % integer(256)).to_vector();
+        if (byte.empty()) byte.push_back(0);
         if (negative)
             byte[0] ^= 255;
         bytes.push_back(byte[0]);
@@ -55,14 +57,24 @@ integer HashPrime(std::vector<uint8_t> seed, int length, vector<int> bitmask) {
     }
 }
 
-std::vector<unsigned char> SerializeForm(form &y, int int_size) {
+std::vector<unsigned char> SerializeForm(form &y, int d_bits)
+{
     y.reduce();
-    std::vector<unsigned char> res = ConvertIntegerToBytes(y.a, int_size);
-    std::vector<unsigned char> b_res = ConvertIntegerToBytes(y.b, int_size);
-    res.insert(res.end(), b_res.begin(), b_res.end());
+    int form_size = bqfc_get_compr_size(d_bits);
+    std::vector<unsigned char> res(form_size);
+    bqfc_serialize(res.data(), y.a.impl, y.b.impl, d_bits);
     return res;
 }
 
+form DeserializeForm(const integer &D, const uint8_t *bytes, size_t size)
+{
+    integer a, b;
+    if (bqfc_deserialize(a.impl, b.impl, D.impl, bytes, size, D.num_bits())) {
+        throw std::runtime_error("Deserializing compressed form failed");
+    }
+    return form::from_abd(a, b, D);
+}
+
 integer FastPow(uint64_t a, uint64_t b, integer& c) {
     integer res, a1 = integer(a);
     mpz_powm_ui(res.impl, a1.impl, b, c.impl);
@@ -70,9 +82,9 @@ integer FastPow(uint64_t a, uint64_t b, integer& c) {
 }
 
 integer GetB(const integer& D, form &x, form& y) {
-    int int_size = (D.num_bits() + 16) >> 4;
-    std::vector<unsigned char> serialization = SerializeForm(x, int_size);
-    std::vector<unsigned char> serialization_y = SerializeForm(y, int_size);
+    int d_bits = D.num_bits();
+    std::vector<unsigned char> serialization = SerializeForm(x, d_bits);
+    std::vector<unsigned char> serialization_y = SerializeForm(y, d_bits);
     serialization.insert(serialization.end(), serialization_y.begin(), serialization_y.end());
     return HashPrime(serialization, 264, {263});
 }
@@ -107,28 +119,27 @@ class PulmarkReducer {
 
 form FastPowFormNucomp(form x, integer &D, integer num_iterations, integer &L, PulmarkReducer& reducer)
 {
-    form res = form::identity(D);
-    int D_size = D.impl->_mp_size;
+    if (!mpz_sgn(num_iterations.impl))
+        return form::identity(D);
 
-    while (1) {
-        if (num_iterations.get_bit(0)) {
+    form res = x;
+    int max_size = -D.impl->_mp_size / 2 + 1, i;
+
+    // Do exponentiation by squaring from top bits of exponent to bottom
+    for (i = num_iterations.num_bits() - 2; i >= 0; i--) {
+        nudupl_form(res, res, D, L);
+        if (res.a.impl->_mp_size > max_size) {
+            // Reduce only when 'a' exceeds a half of the discriminant size by
+            // more than one limb
+            reducer.reduce(res);
+        }
+
+        if (num_iterations.get_bit(i)) {
             nucomp_form(res, res, x, D, L);
-            if (res.a.impl->_mp_size > D_size) {
-                // Reduce only when 'a' has more limbs than D
-                reducer.reduce(res);
-            }
-        }
-
-        num_iterations >>= 1;
-        if (num_iterations == 0) {
-            break;
-        }
-
-        nudupl_form(x, x, D, L);
-        if (x.a.impl->_mp_size > D_size) {
-            reducer.reduce(x);
         }
     }
+
+    reducer.reduce(res);
     return res;
 }
 
