@@ -41,6 +41,7 @@
 
 #include <thread>
 #include <future>
+#include <memory>
 
 #include <chrono>
 #include <condition_variable>
@@ -51,6 +52,9 @@
 #include "fast_storage.h"
 #include <boost/asio.hpp>
 
+#include <atomic>
+#include <optional>
+
 bool warn_on_corruption_in_production=false;
 
 using boost::asio::ip::tcp;
@@ -96,7 +100,9 @@ void repeated_square_original(vdf_original &vdfo, form& f, const integer& D, con
 }
 
 // thread safe; but it is only called from the main thread
-void repeated_square(form f, const integer& D, const integer& L, WesolowskiCallback* weso, FastStorage* fast_storage, bool& stopped) {
+void repeated_square(form f, const integer& D, const integer& L,
+    WesolowskiCallback* weso, FastStorage* fast_storage, std::atomic<bool>& stopped)
+{
     #ifdef VDF_TEST
         uint64 num_calls_fast=0;
         uint64 num_iterations_fast=0;
@@ -233,7 +239,9 @@ void repeated_square(form f, const integer& D, const integer& L, WesolowskiCallb
     #endif
 }
 
-Proof ProveOneWesolowski(uint64_t iters, integer& D, OneWesolowskiCallback* weso, bool& stopped) {
+Proof ProveOneWesolowski(uint64_t iters, integer& D, OneWesolowskiCallback* weso,
+    std::atomic<bool>& stopped)
+{
     while (weso->iterations < iters) {
         this_thread::sleep_for(1s);
     }
@@ -244,24 +252,26 @@ Proof ProveOneWesolowski(uint64_t iters, integer& D, OneWesolowskiCallback* weso
         /*x=*/f,
         /*y=*/weso->result
     );
-    OneWesolowskiProver prover(sg, D, weso->forms, stopped);
+    OneWesolowskiProver prover(sg, D, weso->forms.get(), stopped);
     prover.start();
     while (!prover.IsFinished()) {
         this_thread::sleep_for(1s);
     }
-    int int_size = (D.num_bits() + 16) >> 4;
+    int d_bits = D.num_bits();
     std::vector<unsigned char> y_serialized;
     std::vector<unsigned char> proof_serialized;
-    y_serialized = SerializeForm(weso->result, 129);
+    y_serialized = SerializeForm(weso->result, d_bits);
     form proof_form = prover.GetProof();
-    proof_serialized = SerializeForm(proof_form, int_size);
+    proof_serialized = SerializeForm(proof_form, d_bits);
     Proof proof(y_serialized, proof_serialized);
     proof.witness_type = 0;
     std::cout << "Got simple weso proof: " << proof.hex() << "\n";
     return proof;
 }
 
-Proof ProveTwoWeso(integer& D, form x, uint64_t iters, uint64_t done_iterations, TwoWesolowskiCallback* weso, int depth, bool& stop_signal) {
+Proof ProveTwoWeso(integer& D, form x, uint64_t iters, uint64_t done_iterations,
+    TwoWesolowskiCallback* weso, int depth, std::atomic<bool>& stop_signal)
+{
     integer L=root(-D, 4);
     if (depth == 2) {
         while (!stop_signal && weso->iterations < done_iterations + iters) {
@@ -288,9 +298,9 @@ Proof ProveTwoWeso(integer& D, form x, uint64_t iters, uint64_t done_iterations,
             return Proof();
         form proof = prover.GetProof();
 
-        int int_size = (D.num_bits() + 16) >> 4;
-        std::vector<unsigned char> y_bytes = SerializeForm(y, 129);
-        std::vector<unsigned char> proof_bytes = SerializeForm(proof, int_size);
+        int d_bits = D.num_bits();
+        std::vector<unsigned char> y_bytes = SerializeForm(y, d_bits);
+        std::vector<unsigned char> proof_bytes = SerializeForm(proof, d_bits);
         Proof final_proof=Proof(y_bytes, proof_bytes);
         return final_proof;
     }
@@ -322,18 +332,18 @@ Proof ProveTwoWeso(integer& D, form x, uint64_t iters, uint64_t done_iterations,
     if (stop_signal)
         return Proof();
     form proof = prover.GetProof();
-    
-    int int_size = (D.num_bits() + 16) >> 4;
+
+    int d_bits = D.num_bits();
     Proof final_proof;
     final_proof.y = proof2.y;
     std::vector<unsigned char> proof_bytes(proof2.proof);
     std::vector<unsigned char> tmp = ConvertIntegerToBytes(integer(iterations1), 8);
     proof_bytes.insert(proof_bytes.end(), tmp.begin(), tmp.end());
     tmp.clear();
-    tmp = SerializeForm(y1, int_size);
+    tmp = SerializeForm(y1, d_bits);
     proof_bytes.insert(proof_bytes.end(), tmp.begin(), tmp.end());
     tmp.clear();
-    tmp = SerializeForm(proof, int_size);
+    tmp = SerializeForm(proof, d_bits);
     proof_bytes.insert(proof_bytes.end(), tmp.begin(), tmp.end());
     final_proof.proof = proof_bytes;
     if (depth == 0) {
@@ -360,15 +370,11 @@ class ProverManager {
         }
     }
 
-    ~ProverManager() {
-        delete(main_loop);
-    }
-
     void start() {
-        main_loop = new std::thread([=] {RunEventLoop();});
+        main_loop.emplace(&ProverManager::RunEventLoop, this);
     }
 
-    void stop() {        
+    void stop() {
         {
             std::lock_guard<std::mutex> lk(new_event_mutex);
             stopped = true;
@@ -403,7 +409,7 @@ class ProverManager {
             last_segment_cv.wait(lk, [this, iteration] {
                 if (vdf_iteration >= iteration - iteration % (1 << 16))
                     return true;
-                return stopped;
+                return stopped.load();
             });
             if (stopped) {
                 return Proof();
@@ -417,13 +423,13 @@ class ProverManager {
             // Recalculate everything from the checkpoint, since there is no guarantee the iter didn't arrive late.
             integer L = root(-D, 4);
             PulmarkReducer reducer;
-            form* intermediates = (form*) calloc((iteration % (1 << 16)) / 10 + 100, sizeof(form));
+            std::unique_ptr<form[]> intermediates(new form[(iteration % (1 << 16)) / 10 + 100]);
             for (int i = 0; i < iteration % (1 << 16); i++) {
                 if (i % 10 == 0) {
                     intermediates[i / 10] = y;
                 }
                 nudupl_form(y, y, D, L);
-                reducer.reduce(y);   
+                reducer.reduce(y);
                 if (stopped) {
                     return Proof();
                 }
@@ -434,10 +440,9 @@ class ProverManager {
                 /*x=*/weso->checkpoints[iteration / (1 << 16)],
                 /*y=*/y
             );
-            OneWesolowskiProver prover(sg, D, intermediates, stopped);
+            OneWesolowskiProver prover(sg, D, intermediates.get(), stopped);
             prover.start();
             sg.proof = prover.GetProof();
-            free(intermediates);
             if (stopped) {
                 return Proof();
             }
@@ -450,9 +455,9 @@ class ProverManager {
             proof_cv.wait(lk, [this, iteration] {
                 if (max_proving_iteration >= iteration - iteration % (1 << 16)) 
                     return true;
-                return stopped;
+                return stopped.load();
             });
-            if (stopped)    
+            if (stopped)
                 return Proof();
             int blobs = 0;
             for (int i = segment_count - 1; i >= 0; i--) {
@@ -484,20 +489,20 @@ class ProverManager {
             proof_segments.emplace_back(last_segment);
         }
         // y, proof, [iters1, y1, proof1], [iters2, y2, proof2], ...
-        int int_size = (D.num_bits() + 16) >> 4;
+        int d_bits = D.num_bits();
         std::vector<unsigned char> y_serialized;
         std::vector<unsigned char> proof_serialized;
         // Match ClassGroupElement type from the blockchain.
-        y_serialized = SerializeForm(y, 129);
-        proof_serialized = SerializeForm(proof_segments[proof_segments.size() - 1].proof, int_size);
+        y_serialized = SerializeForm(y, d_bits);
+        proof_serialized = SerializeForm(proof_segments[proof_segments.size() - 1].proof, d_bits);
         for (int i = proof_segments.size() - 2; i >= 0; i--) {
             std::vector<unsigned char> tmp = ConvertIntegerToBytes(integer(proof_segments[i].length), 8);
             proof_serialized.insert(proof_serialized.end(), tmp.begin(), tmp.end());
             tmp.clear();
-            tmp = SerializeForm(proof_segments[i].y, int_size);
+            tmp = SerializeForm(proof_segments[i].y, d_bits);
             proof_serialized.insert(proof_serialized.end(), tmp.begin(), tmp.end());
             tmp.clear();
-            tmp = SerializeForm(proof_segments[i].proof, int_size);
+            tmp = SerializeForm(proof_segments[i].proof, d_bits);
             proof_serialized.insert(proof_serialized.end(), tmp.begin(), tmp.end());
         }
         Proof proof(y_serialized, proof_serialized);
@@ -513,6 +518,11 @@ class ProverManager {
     }
 
     void RunEventLoop() {
+        // this is running in a separate thread. Any member variables it
+        // accesses must be one of:
+        // * protected by a mutex
+        // * owned entirely by this thread
+        // * atomic
         const bool multi_proc_machine = (std::thread::hardware_concurrency() >= 16) ? true : false;
         bool warned = false;
         bool increased_proving = false;
@@ -525,7 +535,7 @@ class ProverManager {
                 lk.unlock();
             }
             if (stopped)
-                return; 
+                return;
             // Check if we can prove the last segment for some iteration.
             vdf_iteration = weso->iterations;
             // VDF running longer than expected, increase proving threads count.
@@ -623,12 +633,12 @@ class ProverManager {
 
             // Check if new segments have arrived, and add them as pending proof.
             for (int i = 0; i < segment_count; i++) {
-                uint64_t sg_length = 1LL << (16 + 2 * i); 
+                uint64_t sg_length = 1LL << (16 + 2 * i);
                 while (last_appended[i] + sg_length <= intermediates_iter) {
                     if (stopped) return ;
                     Segment sg(
-                        /*start=*/last_appended[i], 
-                        /*length=*/sg_length, 
+                        /*start=*/last_appended[i],
+                        /*length=*/sg_length,
                         /*x=*/weso->checkpoints[last_appended[i] / (1 << 16)],
                         /*y=*/weso->checkpoints[(last_appended[i] + sg_length) / (1 << 16)]
                     );
@@ -707,7 +717,7 @@ class ProverManager {
                     if (!stopped) {
                         provers.emplace_back(
                             std::make_pair(
-                                std::make_unique<InterruptableProver>(best, D, weso),                            
+                                std::make_unique<InterruptableProver>(best, D, weso),
                                 best
                             )
                         );
@@ -720,18 +730,18 @@ class ProverManager {
     }
 
   private:
-    bool stopped = false;
+    std::atomic<bool> stopped = false;
     int segment_count;
     // Maximum amount of proving threads running at once.
     int max_proving_threads;
-    std::thread* main_loop;
+    std::optional<std::thread> main_loop;
     FastAlgorithmCallback* weso;
     FastStorage* fast_storage;
     // The discriminant used.
     integer D;
     // Active or paused provers currently running.
     std::vector<std::pair<std::unique_ptr<InterruptableProver>, Segment>> provers;
-    // Vectors of segments needing proving, for each segment length. 
+    // Vectors of segments needing proving, for each segment length.
     std::vector<std::vector<Segment>> pending_segments;
     // For each segment length, remember the endpoint of the last segment marked as pending.
     std::vector<uint64_t> last_appended;
@@ -750,7 +760,7 @@ class ProverManager {
     // Maximum iter that can be proved.
     uint64_t max_proving_iteration = 0;
     // Where the VDF thread is at.
-    uint64_t vdf_iteration = 0;
+    std::atomic<uint64_t> vdf_iteration{0};
     bool proof_done;
     uint64_t intermediates_iter;
 };
