@@ -1,5 +1,6 @@
 #include <boost/asio.hpp>
 #include "vdf.h"
+#include <atomic>
 
 using boost::asio::ip::tcp;
 
@@ -29,7 +30,7 @@ void WriteProof(uint64_t iteration, Proof& result, tcp::socket& sock) {
     std::vector<unsigned char> y_size = ConvertIntegerToBytes(integer(result.y.size()), 8);
     bytes.insert(bytes.end(), y_size.begin(), y_size.end());
     bytes.insert(bytes.end(), result.y.begin(), result.y.end());
-    
+
     // Writes the witness type.
     std::vector<unsigned char> witness_type = ConvertIntegerToBytes(integer(result.witness_type), 1);
     bytes.insert(bytes.end(), witness_type.begin(), witness_type.end());
@@ -47,10 +48,10 @@ void WriteProof(uint64_t iteration, Proof& result, tcp::socket& sock) {
         boost::asio::write(sock, boost::asio::buffer(prefix_bytes, 4));
         boost::asio::write(sock, boost::asio::buffer(str_result.c_str(), str_result.size()));
     }
-    PrintInfo("Sended proof");
+    PrintInfo("Sent proof");
 }
 
-void CreateAndWriteProof(ProverManager& pm, uint64_t iteration, bool& stop_signal, tcp::socket& sock) {
+void CreateAndWriteProof(ProverManager& pm, uint64_t iteration, std::atomic<bool>& stop_signal, tcp::socket& sock) {
     Proof result = pm.Prove(iteration);
     if (stop_signal == true) {
         PrintInfo("Got stop signal before completing the proof!");
@@ -59,7 +60,7 @@ void CreateAndWriteProof(ProverManager& pm, uint64_t iteration, bool& stop_signa
     WriteProof(iteration, result, sock);
 }
 
-void CreateAndWriteProofOneWeso(uint64_t iters, integer& D, OneWesolowskiCallback* weso, bool& stop_signal, tcp::socket& sock) {
+void CreateAndWriteProofOneWeso(uint64_t iters, integer& D, OneWesolowskiCallback* weso, std::atomic<bool>& stop_signal, tcp::socket& sock) {
     Proof result = ProveOneWesolowski(iters, D, weso, stop_signal);
     if (stop_signal) {
         PrintInfo("Got stop signal before completing the proof!");
@@ -68,7 +69,7 @@ void CreateAndWriteProofOneWeso(uint64_t iters, integer& D, OneWesolowskiCallbac
     WriteProof(iters, result, sock);
 }
 
-void CreateAndWriteProofTwoWeso(integer& D, form f, uint64_t iters, TwoWesolowskiCallback* weso, bool& stop_signal, tcp::socket& sock) {
+void CreateAndWriteProofTwoWeso(integer& D, form f, uint64_t iters, TwoWesolowskiCallback* weso, std::atomic<bool>& stop_signal, tcp::socket& sock) {
     Proof result = ProveTwoWeso(D, f, iters, 0, weso, 0, stop_signal);
     if (stop_signal) {
         PrintInfo("Got stop signal before completing the proof!");
@@ -77,6 +78,8 @@ void CreateAndWriteProofTwoWeso(integer& D, form f, uint64_t iters, TwoWesolowsk
     WriteProof(iters, result, sock);
 }
 
+char initial_form_s[100];
+
 void InitSession(tcp::socket& sock) {
     boost::system::error_code error;
 
@@ -87,6 +90,10 @@ void InitSession(tcp::socket& sock) {
     disc_int_size = atoi(disc_size);
     boost::asio::read(sock, boost::asio::buffer(disc, disc_int_size), error);
 
+    char form_size;
+    boost::asio::read(sock, boost::asio::buffer(&form_size, 1), error);
+    boost::asio::read(sock, boost::asio::buffer(initial_form_s, form_size), error);
+
     if (error == boost::asio::error::eof)
         return ; // Connection closed cleanly by peer.
     else if (error)
@@ -102,8 +109,6 @@ void InitSession(tcp::socket& sock) {
         PrintInfo( "=== Warn on corruption enabled ===" );
     }
     assert(is_vdf_test); //assertions should be disabled in VDF_MODE==0
-    init_gmp();
-    allow_integer_constructor=true; //make sure the old gmp allocator isn't used
     set_rounding_mode();
 }
 
@@ -136,7 +141,7 @@ uint64_t ReadIteration(tcp::socket& sock) {
     boost::asio::read(sock, boost::asio::buffer(data, size), error);
     uint64_t iters = 0;
     for (int i = 0; i < size; i++)
-        iters = iters * 10 + data[i] - '0';       
+        iters = iters * 10 + data[i] - '0';
     return iters;
 }
 
@@ -145,19 +150,19 @@ void SessionFastAlgorithm(tcp::socket& sock) {
     try {
         integer D(disc);
         integer L=root(-D, 4);
-        form f=form::generator(D);
         PrintInfo("Discriminant = " + to_string(D.impl));
-
+        form f = DeserializeForm(D, (uint8_t *)initial_form_s, sizeof(initial_form_s));
+        PrintInfo("Initial form: " + to_string(f.a.impl) + " " + to_string(f.b.impl));
         std::vector<std::thread> threads;
         const bool multi_proc_machine = (std::thread::hardware_concurrency() >= 16) ? true : false;
-        WesolowskiCallback* weso = new FastAlgorithmCallback(segments, D, multi_proc_machine);
+        WesolowskiCallback* weso = new FastAlgorithmCallback(segments, D, f, multi_proc_machine);
         FastStorage* fast_storage = NULL;
         if (multi_proc_machine) {
-            fast_storage = new FastStorage((FastAlgorithmCallback*)weso);   
+            fast_storage = new FastStorage((FastAlgorithmCallback*)weso);
         }
-        bool stopped = false;
+        std::atomic<bool> stopped(false);
         std::thread vdf_worker(repeated_square, f, std::ref(D), std::ref(L), weso, fast_storage, std::ref(stopped));
-        ProverManager pm(D, (FastAlgorithmCallback*)weso, fast_storage, segments, thread_count);        
+        ProverManager pm(D, (FastAlgorithmCallback*)weso, fast_storage, segments, thread_count);
         pm.start();
 
         // Tell client that I'm ready to get the challenges.
@@ -204,7 +209,7 @@ void SessionOneWeso(tcp::socket& sock) {
             FinishSession(sock);
             return;
         }
-        bool stopped = false;
+        std::atomic<bool> stopped(false);
         WesolowskiCallback* weso = new OneWesolowskiCallback(D, iter);
         FastStorage* fast_storage = NULL;
         std::thread vdf_worker(repeated_square, f, std::ref(D), std::ref(L), weso, fast_storage, std::ref(stopped));
@@ -225,23 +230,23 @@ void SessionOneWeso(tcp::socket& sock) {
 }
 
 void SessionTwoWeso(tcp::socket& sock) {
-    const int kMaxProcessesAllowed = 3;
+    const int kMaxProcessesAllowed = 100;
     InitSession(sock);
     try {
         integer D(disc);
         integer L=root(-D, 4);
-        form f = form::generator(D);
         PrintInfo("Discriminant = " + to_string(D.impl));
+        form f = DeserializeForm(D, (uint8_t *)initial_form_s, sizeof(initial_form_s));
 
         // Tell client that I'm ready to get the challenges.
         boost::asio::write(sock, boost::asio::buffer("OK", 2));
 
-        bool stopped = false;
-        bool stop_vector[100];
+        std::atomic<bool> stopped(false);
+        std::atomic<bool> stop_vector[100];
         std::vector<std::thread> threads;
         // (iteration, thread_id)
         std::set<std::pair<uint64_t, uint64_t> > seen_iterations;
-        WesolowskiCallback* weso = new TwoWesolowskiCallback(D);
+        WesolowskiCallback* weso = new TwoWesolowskiCallback(D, f);
         FastStorage* fast_storage = NULL;
         std::thread vdf_worker(repeated_square, f, std::ref(D), std::ref(L), weso, fast_storage, std::ref(stopped));
 
@@ -288,7 +293,7 @@ void SessionTwoWeso(tcp::socket& sock) {
                     PrintInfo("Running proving for iter: " + to_string(iters));
                     stop_vector[threads.size()] = false;
                     threads.push_back(std::thread(CreateAndWriteProofTwoWeso, std::ref(D), f, iters,
-                                      (TwoWesolowskiCallback*)weso, std::ref(stop_vector[threads.size()]), 
+                                      (TwoWesolowskiCallback*)weso, std::ref(stop_vector[threads.size()]),
                                       std::ref(sock)));
                     if (threads.size() > kMaxProcessesAllowed) {
                         PrintInfo("Stopping proving for iter: " + to_string(max_iter));
@@ -307,10 +312,9 @@ void SessionTwoWeso(tcp::socket& sock) {
 int gcd_base_bits=50;
 int gcd_128_max_iter=3;
 
-int main(int argc, char* argv[])
+int main(int argc, char* argv[]) try
 {
-  try
-  {
+    init_gmp();
     if (argc != 4)
     {
       std::cerr << "Usage: ./vdf_client <host> <port>\n";
@@ -348,8 +352,9 @@ int main(int argc, char* argv[])
         two_weso = true;
         SessionTwoWeso(s);
     }
-  } catch (std::exception& e) {
+    return 0;
+}
+catch (std::exception& e) {
     std::cerr << "Exception: " << e.what() << "\n";
-  } 
-  return 0;
+    return 1;
 }
